"""
FortiClient EMS Class
"""

import logging
import pickle
import re
from pathlib import Path
from typing import Any

import requests

from fotoobo.exceptions import APIError, GeneralWarning

from .fortinet import Fortinet

log = logging.getLogger("fotoobo")


class FortiClientEMS(Fortinet):
    """
    Represents one FortiClient EMS (digital twin)
    """

    ALLOWED_HTTP_METHODS = ["DELETE", "GET", "PATCH", "POST"]

    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        cookie_path: str = "",
        **kwargs: Any,
    ) -> None:
        """
        Set some initial parameters.

        Args:
            hostname:    The hostname of the FortiClient EMS to connect to
            username:    Username
            password:    Password
            cookie_path: Path to write the cookie files (no cookie if empty)
            **kwargs:    See Fortinet class for available arguments
        """
        super().__init__(hostname, **kwargs)
        self.api_url = f"https://{self.hostname}:{self.https_port}/api/v1"
        self.cookie_path = cookie_path
        self.password: str = password
        self.username: str = username
        self.type: str = "forticlientems"

    def api(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        method: str,
        url: str = "",
        headers: dict[str, str] | None = None,
        params: dict[str, str] | None = None,
        payload: dict[str, Any] | None = None,
        timeout: float | None = None,
    ) -> requests.models.Response:
        """
        API request to a FortiClientEMS device.

        Args:
            method:  Request method from [get, post]
            url:     Rest API URL to request data from
            headers: Additional headers (if needed)
            params:  Dictionary with parameters (if needed)
            payload: JSON body for post requests (if needed)
            timeout: The requests read timeout

        Returns:
            Response from the request
        """
        if not headers:
            headers = self.session.headers  # type: ignore

        return super().api(
            method, url, payload=payload, params=params, timeout=timeout, headers=headers
        )

    def get_version(self) -> str:
        """
        Get the FortiClient EMS version.
        According to Fortinet support the FortiClient EMS version may be read out of the endpoint
        /system/consts/get?system_update_time=1. This endpoint is not (yet) documented. Let's hope
        they support it in case of any issue.

        Returns:
            FortiClient EMS version
        """
        try:
            response = self.api("get", "/system/consts/get?system_update_time=1")

        except APIError as err:
            log.warning("'%s' returned: '%s'", self.hostname, err.message)
            raise GeneralWarning(f"{self.hostname} returned: {err.message}") from err

        try:
            ems_version: str = response.json()["data"]["System"]["VERSION"]

        except KeyError as err:
            log.warning("Did not find any FortiClient EMS version number in response")
            raise GeneralWarning(
                "Did not find any FortiClient EMS version number in response"
            ) from err

        return ems_version

    def login(self) -> int:
        """
        Login to the FortiClientEMS.

        Returns:
            Status code from the FortiClient EMS logon
        """
        status = 401
        cookie = Path(self.cookie_path).expanduser() / f"{self.hostname}.cookie"
        csrf = Path(self.cookie_path).expanduser() / f"{self.hostname}.csrf"

        if self.cookie_path:
            log.debug("Searching cookie and csrf token in '%s'", cookie.parents[0])

            if cookie.is_file() and csrf.is_file():
                log.debug("Cookie and csrf token both exist")
                with cookie.open("rb") as cookie_file:
                    self.session.cookies.update(pickle.load(cookie_file))

                self.session.headers["Referer"] = f"https://{self.hostname}"
                self.session.headers["X-CSRFToken"] = csrf.read_text()

                try:
                    response = self.api("get", "/system/serial_number")
                    if (
                        "retval" in response.json()["result"]
                        and int(response.json()["result"]["retval"]) == 1
                    ):
                        log.debug(
                            "Session with given cookie is valid (status: '%s')",
                            response.status_code,
                        )
                        status = response.status_code

                except APIError as err:
                    log.debug("Session with given cookie is invalid (status: '%s')", err.code)
                    status = err.code

            else:
                log.debug("No cookie or csrf token found for '%s'", self.hostname)

        if status == 401:
            log.debug("Login to '%s'", self.hostname)
            payload = {"name": self.username, "password": self.password}
            response = self.api("post", "/auth/signin", payload=payload)

            if response.status_code == 200:
                self.session.headers["Referer"] = f"https://{self.hostname}"
                if match := re.match(r"csrftoken=(\S+);", response.headers["Set-Cookie"]):
                    csrf_token = match.group(1)
                    self.session.headers["X-CSRFToken"] = csrf_token

                if self.cookie_path:
                    log.debug("Saving cookie for '%s'", self.hostname)
                    try:
                        with cookie.open("wb") as cookie_file:
                            pickle.dump(self.session.cookies, cookie_file)

                    except FileNotFoundError as exc:
                        log.debug(exc)
                        log.warning("Unable to save cookie file '%s'", str(cookie.resolve()))

                    log.debug("Saving csrf token for '%s'", self.hostname)
                    try:
                        csrf.write_text(csrf_token)

                    except NameError as exc:
                        log.debug(exc)
                        log.warning("Unable to save csrf token file '%s'", str(csrf.resolve()))

            status = response.status_code

        return status

    def logout(self) -> int:
        """
        Logout from FortiClient EMS.

        Returns:
            Status code from the FortiClient EMS logout
        """
        response = self.api("get", "/auth/signout")
        log.debug("Logged out from '%s' (status: '%s')", self.hostname, response.status_code)
        return response.status_code



"""
The FortiCloudAsset get commands
"""

import logging
from pathlib import Path
from typing import Any, Annotated

import typer

from fotoobo.exceptions.exceptions import GeneralError
from fotoobo.helpers import cli_path
from fotoobo.helpers.files import save_json_file
from fotoobo.helpers.result import Result
from fotoobo.tools import fcasset

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


HELP_TEXT_OPTION_OUTPUT_FILE = "The file to write the output to."


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fcasset get subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command()
def products(
    output_file: Annotated[
        Path | None,
        typer.Option(
            "--output",
            "-o",
            help=HELP_TEXT_OPTION_OUTPUT_FILE,
            metavar="[output]",
        ),
    ] = None,
    raw: Annotated[bool, typer.Option("-r", "--raw", help="Output raw data.")] = False,
) -> None:
    """
    Get the FortiCloud products from asset management.
    """
    result = fcasset.get.products("forticloudasset")
    messages = result.get_messages("forticloudasset")
    if messages:
        for message in messages:
            if message["level"] == "error":
                log.error(message["message"])

            else:
                log.info(message["message"])

        raise GeneralError("Error getting FortiCloud products")

    if output_file:
        log.debug("output_file is: '%s'", output_file)
        save_json_file(output_file, result.get_result("forticloudasset"))

    else:
        if raw:
            result.print_raw()

        else:
            data = Result[Any]()
            for asset in result.get_result("forticloudasset"):
                asset.pop("entitlements")
                asset.pop("contracts")
                asset.pop("productModelEoR")
                asset.pop("productModelEoS")
                asset.pop("accountId")
                asset.pop("folderId")
                asset.pop("assetGroups")
                data.push_result(asset["serialNumber"], asset)

            data.print_result_as_table(title="FortiCloud API products", auto_header=True)


@app.command()
def version() -> None:
    """
    Get the FortiCloud API version.
    """
    result = fcasset.get.version("forticloudasset")
    result.print_result_as_table(
        title="FortiCloud Asset Management API Version",
        headers=["FortiCloud Asset Management", "Version"],
    )



"""
The fotoobo convert commands
"""

import logging
from pathlib import Path
from typing import Annotated

import typer

from fotoobo.helpers import cli_path
from fotoobo.helpers.files import load_json_file, save_json_file
from fotoobo.tools import convert

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fotoobo convert command callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command(no_args_is_help=True)
def checkpoint(
    infile: Annotated[
        Path,
        typer.Argument(
            help="The json file to read the Checkpoint objects from.",
            show_default=False,
            metavar="[infile]",
        ),
    ],
    outfile: Annotated[
        Path,
        typer.Argument(
            help="The json file to write the converted objects to.",
            show_default=False,
            metavar="[outfile]",
        ),
    ],
    obj_type: Annotated[
        str,
        typer.Argument(
            help="The type of objects to convert.",
            show_default=False,
            metavar="[type]",
        ),
    ],
    cache_dir: Annotated[
        Path | None,
        typer.Argument(
            help="The cache directory to use.",
            show_default=False,
            metavar="[cache_dir]",
        ),
    ] = None,
) -> None:
    """
    Convert Checkpoint assets into Fortinet objects.

    The Checkpoint objects have to be prepared in a json file. See
    https://fotoobo.readthedocs.io/en/latest/usage/convert.html for more information.
    """
    checkpoint_assets = load_json_file(infile)
    result = convert.checkpoint(checkpoint_assets, obj_type, outfile.name, cache_dir)
    save_json_file(outfile, result.get_result("fortinet_assets"))



"""
FortiGate CMDB firewall service group module
"""

from pathlib import Path
from typing import Any

from fotoobo.fortinet.fortigate import FortiGate
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory


def get_cmdb_firewall_service_group(
    host: str, name: str, vdom: str, output_file: str | None
) -> Result[list[Any]]:
    """
    Get the firewall service group object(s).

    The FortiGate api endpoint is: /cmdb/firewall.service/group
    """
    inventory = Inventory(config.inventory_file)
    fgt: FortiGate = inventory.get_item(host, "fortigate")
    result = Result[list[Any]]()

    service_group_list = fgt.api_get(url=f"/cmdb/firewall/addrgrp/{name}", vdom=vdom)

    if output_file:
        result.push_result(key=host, data=service_group_list)
        result.save_raw(file=Path(output_file), key=host)

    assets = []
    if service_group_list:
        for vd in service_group_list:
            for asset in vd["results"]:
                data: dict[str, str] = {
                    "name": asset["name"],
                    "vdom": vd["vdom"],
                    "content": "\n".join(_["name"] for _ in asset["member"]),
                }

                assets.append(data)

    result.push_result(host, assets)

    return result



"""
The FortiGate commands
"""

import logging
import os
from datetime import datetime
from pathlib import Path
from typing import Annotated

import typer

from fotoobo import tools
from fotoobo.exceptions import GeneralWarning
from fotoobo.helpers import cli_path
from fotoobo.helpers.config import config as fotoobo_config
from fotoobo.helpers.files import create_dir, file_to_ftp, file_to_zip
from fotoobo.inventory import Inventory

from . import config, monitor
from .get import get

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


app.add_typer(config.app, name="config", help="FortiGate config file commands.")
app.add_typer(get.app, name="get", help="FortiGate get commands.")
app.add_typer(monitor.app, name="monitor", help="FortiGate monitor commands.")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fgt subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command()
def backup(
    host: Annotated[
        str,
        typer.Argument(
            help="The FortiGate to backup (must be defined in the inventory). "
            "Backups all if left empty.",
            show_default=False,
            metavar="[host]",
        ),
    ] = "",
    timeout: Annotated[
        int,
        typer.Option(
            "--timeout",
            "-t",
            help="Timeout in seconds to wait for each FortiGate to respond.",
            show_default=True,
        ),
    ] = 60,
    backup_dir: Annotated[
        Path | None,
        typer.Option(
            "--backup-dir",
            "-b",
            help="Directory to save the backup(s) to. Default is the current working directory.",
            show_default=False,
            metavar="backup_dir",
        ),
    ] = None,
    ftp_server: Annotated[
        str | None,
        typer.Option(
            "--ftp",
            "-f",
            help="The ftp configuration from the inventory to send the backup to.",
            metavar="server",
        ),
    ] = None,
    smtp_server: Annotated[
        str | None,
        typer.Option(
            "--smtp",
            "-s",
            help="The smtp configuration from the inventory to send potential errors to.",
            metavar="server",
        ),
    ] = None,
) -> None:
    """
    Backup one or more FortiGate(s).
    """
    inventory = Inventory(fotoobo_config.inventory_file)

    if not backup_dir:
        backup_dir = Path.cwd()

    create_dir(backup_dir)
    result = tools.fgt.backup(host, timeout=timeout)

    for name, data in result.all_results().items():
        config_file = backup_dir / Path(name).with_suffix(".conf")

        if config_file.is_file():
            os.remove(config_file)

        config_file.write_text(data, encoding="UTF-8")

        if not config_file.is_file():
            result.push_message(name, f"backup file for '{name}' does not exist")
            continue

        if ftp_server:
            if ftp_server in inventory.assets:
                server = inventory.assets[ftp_server]
                log.debug("Compressing configuration '%s'", name)
                time: str = datetime.now().strftime("%Y%m%d-%H%M")
                zip_file = backup_dir / Path(name + "-" + time + ".conf.zip")
                file_to_zip(config_file, zip_file)
                file_to_ftp(zip_file, server)
                os.remove(zip_file)
                result.push_message(name, f"Uploaded config file for '{name}' to '{ftp_server}'")

            else:
                raise GeneralWarning(f"FTP server '{ftp_server}' not found in inventory")

    if smtp_server and smtp_server in inventory.assets:
        result.send_messages_as_mail(inventory.assets[smtp_server], "error")



"""
FortiCloud cloud asset get utility
"""

import logging
from typing import Any

from fotoobo.fortinet.forticloudasset import FortiCloudAsset
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory

log = logging.getLogger("fotoobo")


def products(host: str) -> Result[list[dict[str, Any]]]:
    """
    FortiCloud API get products from Asset Management

    Here we assume that just one page is returned. If it happens that more that one page is returned
    pagination has to be implemented.

    Args:
        host: Host defined in inventory

    Returns:
        The asset list as a result object
    """
    result = Result[list[dict[str, Any]]]()
    inventory = Inventory(config.inventory_file)
    fc: FortiCloudAsset = inventory.get_item(host, "forticloudasset")
    log.debug("FortiCloud get assets ...")
    response = fc.post("/products/list", payload={"serialNumber": "%"})
    if response["error"]:
        result.push_message(host, response["message"], "info")

    else:
        result.push_result(host, response["assets"])

    return result


def version(host: str) -> Result[str]:
    """
    FortiCloud API get version

    Args:
        host: Host defined in inventory

    Returns:
        The version string for the FortiCloud API in a result object
    """
    result = Result[str]()
    inventory = Inventory(config.inventory_file)
    fcasset: FortiCloudAsset = inventory.get_item(host, "forticloudasset")
    log.debug("FortiCloud Asset Management get version ...")
    result.push_result(host, fcasset.get_version())

    return result



"""
FortiGate CMDB firewall service custom module
"""

from pathlib import Path
from typing import Any

from fotoobo.fortinet.fortigate import FortiGate
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory


def get_cmdb_firewall_service_custom(
    host: str, name: str, vdom: str, output_file: str | None
) -> Result[list[Any]]:
    """
    Get the firewall service custom object(s).

    The FortiGate api endpoint is: /cmdb/firewall.service/custom
    """
    inventory = Inventory(config.inventory_file)
    fgt: FortiGate = inventory.get_item(host, "fortigate")
    result = Result[list[Any]]()

    service_custom_list = fgt.api_get(url=f"/cmdb/firewall.service/custom/{name}", vdom=vdom)

    if output_file:
        result.push_result(key=host, data=service_custom_list)
        result.save_raw(file=Path(output_file), key=host)

    assets = []
    if service_custom_list:
        for vd in service_custom_list:
            for asset in vd["results"]:

                data: dict[str, str] = {
                    "name": asset["name"],
                    "vdom": vd["vdom"],
                    "protocol": asset["protocol"],
                }

                if asset["protocol"] == "TCP/UDP/SCTP":
                    data["data_1"] = asset.get("tcp-portrange", "")
                    data["data_2"] = asset.get("udp-portrange", "")

                elif asset["protocol"] in ["ICMP", "ICMP6"]:
                    data["data_1"] = asset.get("icmptype", "")
                    data["data_2"] = asset.get("icmpcode", "")

                elif asset["protocol"] == "IP":
                    data["data_1"] = asset.get("protocol-number", "")
                    data["data_2"] = ""

                else:
                    data["data_1"] = ""
                    data["data_2"] = ""

                assets.append(data)

    result.push_result(host, assets)

    return result



"""
FortiGate CMDB firewall addrgrp module
"""

from pathlib import Path
from typing import Any

from fotoobo.fortinet.fortigate import FortiGate
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory


def get_cmdb_firewall_addrgrp(
    host: str, name: str, vdom: str, output_file: str | None
) -> Result[list[Any]]:
    """
    Get the firewall address group object(s).

    The FortiGate api endpoint is: /cmdb/firewall/addrgrp
    """
    inventory = Inventory(config.inventory_file)
    fgt: FortiGate = inventory.get_item(host, "fortigate")
    result = Result[list[Any]]()

    addrgrp_list = fgt.api_get(url=f"/cmdb/firewall/addrgrp/{name}", vdom=vdom)

    if output_file:
        result.push_result(key=host, data=addrgrp_list)
        result.save_raw(file=Path(output_file), key=host)

    assets = []
    if addrgrp_list:
        for vd in addrgrp_list:
            for asset in vd["results"]:
                # print(asset)
                data: dict[str, str] = {
                    "name": asset["name"],
                    "vdom": vd["vdom"],
                    "content": "\n".join(_["name"] for _ in asset["member"]),
                }

                assets.append(data)

    result.push_result(host, assets)

    return result



"""
fgt tools cmdb firewall
"""

from .address import get_cmdb_firewall_address
from .addrgrp import get_cmdb_firewall_addrgrp
from .service_custom import get_cmdb_firewall_service_custom
from .service_group import get_cmdb_firewall_service_group

__all__ = [
    "get_cmdb_firewall_address",
    "get_cmdb_firewall_addrgrp",
    "get_cmdb_firewall_service_custom",
    "get_cmdb_firewall_service_group",
]



"""
FortiCloud Class
"""

import logging
from pathlib import Path
from typing import Any

import requests

from fotoobo.exceptions.exceptions import APIError, GeneralWarning

from .fortinet import Fortinet

log = logging.getLogger("fotoobo")


class FortiCloudAsset(Fortinet):
    """
    Represents the FortiCloud Asset Management (digital twin)

    Helpful documentation links:
    - https://fndn.fortinet.net/index.php?/fortiapi/55-asset-management-formerly-forticare-registration/4638/55/Folder/ # pylint: disable=line-too-long
    - https://docs.fortinet.com/document/forticloud/25.2.0/identity-access-management-iam/19322/accessing-fortiapis # pylint: disable=line-too-long
    """

    ALLOWED_HTTP_METHODS = ["POST"]

    def __init__(self, username: str, password: str, **kwargs: Any) -> None:
        """
        Set some initial parameters.

        Args:
            username: The username
            password: The password

        Keyword Args:
            token_path: The path where to load/save the access token
            **kwargs:  See Fortinet class for more available arguments
        """
        super().__init__("support.fortinet.com", **kwargs)
        self.api_url = f"https://{self.hostname}:{self.https_port}/ES/api/registration/v3"
        self.password = password
        self.username = username
        self.access_token: str = ""
        self.token_path: str = kwargs.get("token_path", "")
        self.type = "forticloudasset"

    def api(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        method: str,
        url: str = "",
        headers: dict[str, str] | None = None,
        params: dict[str, str] | None = None,
        payload: dict[str, Any] | None = None,
        timeout: float | None = None,
    ) -> requests.models.Response:
        """
        API request to a FortiManager device.

        It uses the super.api method but it has to enrich the payload in post requests with the
        needed session key.

        Args:
            method:  Request method from [get, post]
            url:     Rest API URL to request data from
            headers: Dictionary with headers (if needed)
            params:  Dictionary with parameters (if needed)
            payload: JSON body for post requests (if needed)
            timeout: The requests read timeout in seconds

        Returns:
            Response: Response object from the request
        """
        if not self.access_token:
            self.login()

        payload = payload or {}
        headers = headers or {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.access_token}",
        }
        return super().api(
            method, url, headers=headers, payload=payload, params=params, timeout=timeout
        )

    def get_version(self) -> str:
        """
        Get FortiCloud API version.

        Returns:
            FortiCloud API version
        """
        fc_version: str = ""

        try:
            fc_version = self.post(url="/folders/list")["version"]

        except APIError as err:
            log.warning("'%s' returned: '%s'", self.hostname, err)
            raise GeneralWarning(f"{self.hostname} returned: {err}") from err

        return fc_version

    def login(self) -> int:
        """
        Login to the FortiCloud.

        If the login to the FortiCloud was successful it stores the access token in the object.
        We do not use requests.session as the access token is just a string which is saved directly.

        Returns:
            Status code from the FortiCloud login
        """
        status: int = 401

        if self.token_path:
            token_file = Path(self.token_path).expanduser() / f"{self.hostname}.token"
            if token_file.exists():
                log.debug("Loading access token from file '%s'", token_file)

                with token_file.open(encoding="UTF-8") as file:
                    self.access_token = file.read()

                headers = {
                    "Content-Type": "application/json",
                    "Authorization": f"Bearer {self.access_token}",
                }
                try:
                    response = requests.post(
                        url=f"{self.api_url}/folders/list",
                        headers=headers,
                        verify=self.ssl_verify,
                        timeout=10,
                    )
                    status = response.status_code

                    if response.status_code == 200:
                        log.debug("Access token is still valid")

                    else:
                        self.access_token = ""
                        log.debug("Access token is invalid")

                except ValueError:
                    self.access_token = ""
                    log.debug("Access token is invalid")

            else:
                log.debug("Token file '%s' does not exist", token_file)

        if not self.access_token and self.username and self.password:
            log.debug("Login to '%s'", self.hostname)
            url = "https://customerapiauth.fortinet.com/api/v1/oauth/token/"
            payload = {
                "username": self.username,
                "password": self.password,
                "client_id": "assetmanagement",
                "grant_type": "password",
            }
            headers = {"Content-Type": "application/json"}
            response = requests.post(
                url, headers=headers, json=payload, verify=self.ssl_verify, timeout=10
            )

            if response.status_code == 200:
                if "access_token" in response.json():
                    self.access_token = response.json()["access_token"]
                    if self.token_path:
                        log.debug("Saving access token into file '%s'", token_file)

                        with token_file.open("w", encoding="UTF-8") as file:
                            file.write(self.access_token)

            status = response.status_code

        return status

    def post(self, url: str, payload: dict[str, Any] | None = None) -> Any:
        """
        POST method to FortiCloud.

        You can pass a single payload (dict) or a list of payloads (list of dict).

        Args:
            payload: The payload to send with the request

        Returns:
            The result of the request
        """
        payload = payload or {}
        response = self.api("post", url=url, payload=payload, timeout=10)

        return response.json()



"""
This is the main init file for the cli module

Do not add too much logic into the cli commands. Just add the tools module and add the
logic there. Therefore, we can segregate the duties.
"""

from . import convert, get

__all__ = ["convert", "get"]



"""
The FortiManager commands
"""

import logging
from pathlib import Path
from typing import Annotated

import typer

from fotoobo.helpers import cli_path
from fotoobo.helpers.config import config as fotoobo_config
from fotoobo.inventory import Inventory
from fotoobo.tools import fmg

from . import get

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fmg subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command(no_args_is_help=True)
def assign(
    adoms: Annotated[
        str,
        typer.Argument(
            help="The ADOMs to assign the global policy/objects to. Use 'fotoobo fmg get adoms' to "
            "get a list of available ADOMs. Separate multiple ADOMs by comma (no spaces).",
            metavar="[adoms]",
            show_default=False,
        ),
    ],
    policy: Annotated[
        str,
        typer.Argument(
            help="The global policy to assign",
            metavar="[policy]",
            show_default=False,
        ),
    ],
    host: Annotated[
        str,
        typer.Argument(
            help="The FortiManager to access (must be defined in the inventory).",
            metavar="[host]",
        ),
    ] = "fmg",
    smtp_server: Annotated[
        str | None,
        typer.Option(
            "--smtp",
            "-s",
            help="The smtp configuration from the inventory to send potential errors to.",
            metavar="[server]",
        ),
    ] = None,
    timeout: Annotated[
        int,
        typer.Option(
            "--timeout",
            "-t",
            help="The timeout to wait for the FortiManager task to finish.",
            metavar="[timeout]",
        ),
    ] = 60,
) -> None:
    """
    Assign a global policy to a specified ADOM or to a list of ADOMs.
    """
    inventory = Inventory(fotoobo_config.inventory_file)
    result = fmg.assign(adoms=adoms, policy=policy, host=host, timeout=timeout)

    if smtp_server:
        if smtp_server in inventory.assets:
            result.send_messages_as_mail(inventory.assets[smtp_server], "error")

        else:
            log.warning("SMTP server '%s' not in found in inventory.", smtp_server)


@app.command(no_args_is_help=True)
def post(
    file: Annotated[
        Path,
        typer.Argument(help="JSON file with payload(s).", show_default=False, metavar="[file]"),
    ],
    adom: Annotated[
        str,
        typer.Argument(
            help="The ADOM to issue the set command(s).", metavar="[adom]", show_default=False
        ),
    ],
    host: Annotated[
        str,
        typer.Argument(
            help="The FortiManager to access (must be defined in the inventory).",
            metavar="[host]",
        ),
    ] = "fmg",
    smtp_server: Annotated[
        str | None,
        typer.Option(
            "--smtp",
            "-s",
            help="The smtp configuration from the inventory to send potential errors to.",
            metavar="[server]",
        ),
    ] = None,
) -> None:
    """
    Post any valid JSON request to the FortiManager.

    Configure the FortiManager with any valid API call(s) given within the JSON file.
    """
    inventory = Inventory(fotoobo_config.inventory_file)
    result = fmg.post(file=file, adom=adom, host=host)

    if smtp_server:
        if smtp_server in inventory.assets:
            result.send_messages_as_mail(inventory.assets[smtp_server], "error", command=True)

        else:
            log.warning("SMTP server '%s' not in found in inventory.", smtp_server)


app.add_typer(get.app, name="get", help="FortiManager get commands.")



"""
FortiGate CMDB firewall address module
"""

from pathlib import Path
from typing import Any


from fotoobo.fortinet.fortigate import FortiGate
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory


def get_cmdb_firewall_address(
    host: str, name: str, vdom: str, output_file: str | None
) -> Result[list[Any]]:
    """
    Get the firewall address object(s).

    The FortiGate api endpoint is: /cmdb/firewall/address
    """
    inventory = Inventory(config.inventory_file)
    fgt: FortiGate = inventory.get_item(host, "fortigate")
    result = Result[list[Any]]()

    address_list = fgt.api_get(url=f"/cmdb/firewall/address/{name}", vdom=vdom)
    result.push_result(key=host, data=address_list)

    if output_file:
        result.push_result(key=host, data=address_list)
        result.save_raw(file=Path(output_file), key=host)

    assets = []
    if address_list:
        for vd in address_list:
            for asset in vd["results"]:

                data: dict[str, str] = {
                    "name": asset["name"],
                    "vdom": vd["vdom"],
                    "type": asset["type"],
                }

                if asset["type"] == "fqdn":
                    data["content"] = asset["fqdn"]

                elif asset["type"] == "geography":
                    data["content"] = asset["country"]

                elif asset["type"] == "ipmask":
                    data["content"] = "/".join(
                        [asset["subnet"].split(" ")[0], asset["subnet"].split(" ")[1]]
                    )

                elif asset["type"] == "iprange":
                    data["content"] = " - ".join([asset["start-ip"], asset["end-ip"]])

                else:
                    data["content"] = ""

                assets.append(data)

    result.push_result(host, assets)

    return result



"""
The FortiGate commands
"""

import logging
from typing import Annotated

import typer

from fotoobo.exceptions import GeneralError
from fotoobo.helpers import cli_path
from fotoobo.tools.fgt.cmdb.firewall import *  # pylint: disable=wildcard-import

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fgt get cmdb get firewall subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command()
def address(
    host: Annotated[
        str,
        typer.Argument(
            help="The FortiGate hostname to access (must be defined in the inventory).",
            metavar="[host]",
        ),
    ],
    name: Annotated[
        str,
        typer.Argument(
            help="The firewall address object to get.",
            metavar="[name]",
        ),
    ] = "",
    vdom: Annotated[
        str,
        typer.Option(
            "--vdom",
            help="The vdom to query ('vdom1' or 'vdom1,vdom2').",
            metavar="[vdom]",
        ),
    ] = "*",
    output_file: Annotated[
        str | None,
        typer.Option(
            "--output",
            "-o",
            help="Output file (format is specified by extension).",
            metavar="[file]",
            show_default=False,
        ),
    ] = None,
) -> None:
    """Get FortiGate cmdb firewall address."""
    if name and ("*" in vdom or "," in vdom):
        raise GeneralError("With name argument you have to specify one single VDOM (with --vdom)")

    result = get_cmdb_firewall_address(host, name, vdom, output_file)
    if not output_file:
        result.print_table_raw(result.results[host], auto_header=True, title=host)


@app.command()
def addrgrp(
    host: Annotated[
        str,
        typer.Argument(
            help="The FortiGate hostname to access (must be defined in the inventory).",
            metavar="[host]",
        ),
    ],
    name: Annotated[
        str,
        typer.Argument(
            help="The firewall address group object to get.",
            metavar="[name]",
        ),
    ] = "",
    vdom: Annotated[
        str,
        typer.Option(
            "--vdom",
            help="The vdom to query ('vdom1' or 'vdom1,vdom2').",
            metavar="[vdom]",
        ),
    ] = "*",
    output_file: Annotated[
        str | None,
        typer.Option(
            "--output",
            "-o",
            help="Output file (format is specified by extension).",
            metavar="[file]",
            show_default=False,
        ),
    ] = None,
) -> None:
    """Get FortiGate cmdb firewall address group."""
    if name and ("*" in vdom or "," in vdom):
        raise GeneralError("With name argument you have to specify one single VDOM (with --vdom)")

    result = get_cmdb_firewall_addrgrp(host, name, vdom, output_file)
    if not output_file:
        result.print_table_raw(result.results[host], auto_header=True, title=host)


@app.command()
def service_custom(
    host: Annotated[
        str,
        typer.Argument(
            help="The FortiGate hostname to access (must be defined in the inventory).",
            metavar="[host]",
        ),
    ],
    name: Annotated[
        str,
        typer.Argument(
            help="The firewall service object to get.",
            metavar="[name]",
        ),
    ] = "",
    vdom: Annotated[
        str,
        typer.Option(
            "--vdom",
            help="The vdom to query ('vdom1' or 'vdom1,vdom2').",
            metavar="[vdom]",
        ),
    ] = "*",
    output_file: Annotated[
        str | None,
        typer.Option(
            "--output",
            "-o",
            help="Output file (format is specified by extension).",
            metavar="[file]",
            show_default=False,
        ),
    ] = None,
) -> None:
    """Get FortiGate cmdb firewall service custom."""
    if name and ("*" in vdom or "," in vdom):
        raise GeneralError("With name argument you have to specify one single VDOM (with --vdom)")

    result = get_cmdb_firewall_service_custom(host, name, vdom, output_file)
    if not output_file:
        result.print_table_raw(result.results[host], auto_header=True, title=host)


@app.command()
def service_group(
    host: Annotated[
        str,
        typer.Argument(
            help="The FortiGate hostname to access (must be defined in the inventory).",
            metavar="[host]",
        ),
    ],
    name: Annotated[
        str,
        typer.Argument(
            help="The firewall service group to get.",
            metavar="[name]",
        ),
    ] = "",
    vdom: Annotated[
        str,
        typer.Option(
            "--vdom",
            help="The vdom to query ('vdom1' or 'vdom1,vdom2').",
            metavar="[vdom]",
        ),
    ] = "*",
    output_file: Annotated[
        str | None,
        typer.Option(
            "--output",
            "-o",
            help="Output file (format is specified by extension).",
            metavar="[file]",
            show_default=False,
        ),
    ] = None,
) -> None:
    """Get FortiGate cmdb firewall service group."""
    if name and ("*" in vdom or "," in vdom):
        raise GeneralError("With name argument you have to specify one single VDOM (with --vdom)")

    result = get_cmdb_firewall_service_group(host, name, vdom, output_file)
    if not output_file:
        result.print_table_raw(result.results[host], auto_header=True, title=host)



"""
The FortiGate commands
"""

import logging

import typer

from fotoobo.helpers import cli_path

from .firewall import firewall

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


app.add_typer(firewall.app, name="firewall", help="FortiGate get cmdb firewall commands.")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fgt get cmdb subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)



"""
This is the main cli module. If fotoobo is started from the command line the main script starts
app() from within here. Every level of the command line tool is represented by a subdirectory.
If you extend the cli make sure you are in the correct directory or create a new one if there are
more than a few subcommands.

Caution: Use docstrings with care as they are used to print help texts on any command.
"""

import logging
import os
import sys
from pathlib import Path
from typing import Annotated

import typer

from fotoobo import tools
from fotoobo.helpers import cli_path
from fotoobo.helpers.config import config
from fotoobo.helpers.log import Log
from fotoobo.helpers.output import print_logo

from . import convert, get
from .cloud import cloud
from .ems import ems
from .faz import faz
from .fgt import fgt
from .fmg import fmg

app = typer.Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
    rich_markup_mode="rich",
)
log = logging.getLogger("fotoobo")


def version_callback(value: bool) -> None:
    """The version callback"""
    if value:
        if not config.no_logo:
            print_logo()

        get.version(False)
        raise typer.Exit()


@app.callback()
def callback(  # pylint: disable=too-many-arguments, too-many-positional-arguments
    context: typer.Context,
    config_file: Annotated[
        Path | None,
        typer.Option(
            "--config",
            "-c",
            help="Set the fotoobo configuration file.",
            show_default=False,
            metavar="[file]",
        ),
    ] = None,
    log_level: Annotated[
        str | None,
        typer.Option(
            "--loglevel",
            help="Set the log level. Choose from CRITICAL, ERROR, WARNING, INFO, DEBUG.",
            metavar="[level]",
            show_default=False,
        ),
    ] = None,
    nologo: Annotated[
        bool | None, typer.Option("--nologo", help="Do not print the beautiful fotoobo logo.")
    ] = None,
    log_quiet: Annotated[
        bool, typer.Option("--quiet", "-q", help="Disable console logging.", show_default=False)
    ] = False,
    version: Annotated[  # pylint: disable=unused-argument
        bool | None,
        typer.Option(
            "--version", "-V", help="Print the fotoobo version.", callback=version_callback
        ),
    ] = None,
) -> None:
    """
    The Fortinet Toolbox (fotoobo) - make IT easy

    ─────────────────────────────────────────────

    This is fotoobo, the mighty [bold]Fo[/bold]rtinet [bold]too[/bold]l[bold]bo[/bold]x for
    managing your Fortinet environment. It is meant to be extendable to your needs.

    For detailed documentation see https://fotoobo.readthedocs.io.
    """
    config.load_configuration(config_file)
    config.no_logo = True if nologo else config.no_logo

    if log_level:
        log_level = log_level.upper()

    Log.configure_logging(log_quiet, log_level)

    if not config.no_logo:
        print_logo()

    log.debug("Let the magic begin")

    # Log all the configuration options
    # If a config option is a data-structure (dict) unpack it and write it line-by-line
    # For security reasons all sensitive values are shortened
    for attr in dir(config):
        if attr.startswith("_") or attr in ["config", "load_configuration"]:
            continue

        if attr in ["audit_logging", "logging", "vault"] and getattr(config, attr):
            for sub_attr, value in getattr(config, attr).items():
                if attr == "vault" and sub_attr in ["role_id", "secret_id"]:
                    value = f"{value[:4]}...{value[-4:]}"

                log.debug("Option '%s.%s' is '%s'", attr, sub_attr, value)

            continue

        log.debug("Option '%s' is '%s'", attr, getattr(config, attr))

    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)
    log.audit(f'command="{" ".join(sys.argv)}"')  # type: ignore
    config.cli_info = context.to_info_dict()


@app.command(hidden=True)
def greet(
    name: Annotated[
        str | None,
        typer.Argument(
            help="The name of the person to greet.", show_default=False, metavar="[name]"
        ),
    ] = None,
    bye: Annotated[bool, typer.Option("--bye", "-b", help='Also write "bye" at the end.')] = False,
    log_enabled: Annotated[bool, typer.Option("--log", "-l", help="Enable logging.")] = False,
) -> None:
    """
    This is the hidden greeting function.
    It allows you to greet someone with different colors in different languages.
    """
    if not name:
        try:
            name = os.getlogin().capitalize()

        except OSError:  # We need this, will fail on GitHub otherwise...
            name = ""

    tools.greet(str(name), bye, log_enabled)


# fotoobo specific commands
app.add_typer(convert.app, name="convert", help="Convert commands for fotoobo.")
app.add_typer(get.app, name="get", help="Get information about fotoobo or your configuration.")

# commands for the Fortinet products
app.add_typer(ems.app, name="ems", help="Commands for FortiClient EMS.")
app.add_typer(faz.app, name="faz", help="Commands for FortiAnalyzer.")
app.add_typer(cloud.app, name="cloud", help="Commands for FortiCloud applications.")
app.add_typer(fgt.app, name="fgt", help="Commands for FortiGate.")
app.add_typer(fmg.app, name="fmg", help="Commands for FortiManager.")



"""
The FortiGate get commands
"""

import logging
from pathlib import Path
from typing import Annotated

import typer

from fotoobo.helpers.config import config
from fotoobo.inventory.inventory import Inventory
from fotoobo.tools import fgt

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


@app.command(no_args_is_help=True)
def check(
    configuration: Annotated[
        Path,
        typer.Argument(
            help="The FortiGate configuration file or directory.",
            metavar="[config]",
            show_default=False,
        ),
    ],
    bundles: Annotated[
        Path,
        typer.Argument(
            help="Filename of the file containing the check bundles.",
            metavar="[bundles]",
            show_default=False,
        ),
    ],
    smtp_server: Annotated[
        str | None,
        typer.Option(
            "--smtp",
            help="The smtp configuration from the inventory.",
            metavar="[server]",
            show_default=False,
        ),
    ] = None,
) -> None:
    """
    Check one or more FortiGate configuration files.
    """
    inventory = Inventory(config.inventory_file)
    result = fgt.config.check(configuration, bundles)

    if smtp_server:
        if smtp_server in inventory.assets:
            result.send_messages_as_mail(
                inventory.assets[smtp_server],
                count=True,
                command=True,
            )

        else:
            log.warning("SMTP server '%s' not in found in inventory.", smtp_server)

    result.print_messages()


@app.command(no_args_is_help=True)
def get(
    configuration: Annotated[
        Path,
        typer.Argument(
            help="The FortiGate configuration file or directory.",
            metavar="[config]",
            show_default=False,
        ),
    ],
    scope: Annotated[
        str,
        typer.Argument(help="Scope of the configuration ('global' or 'vdom')", metavar="[scope]"),
    ],
    path: Annotated[str, typer.Argument(help="Configuration path", metavar="[path]")] = "/",
) -> None:
    """Get configuration or parts of it from one or more FortiGate configuration files."""
    result = fgt.config.get(configuration, scope, path)
    result.print_raw()


@app.command(no_args_is_help=True)
def info(
    configuration: Annotated[
        Path,
        typer.Argument(
            help="The FortiGate configuration file or directory.",
            metavar="[config]",
            show_default=False,
        ),
    ],
    as_list: Annotated[
        bool,
        typer.Option("--list", "-l", help="Print the result as a list instead of separate blocks."),
    ] = False,
) -> None:
    """
    Get the information from one or more FortiGate configuration files.
    """
    result = fgt.config.info(configuration)

    if as_list:
        info_dicts = []
        for _, _info in result.all_results().items():
            info_dicts.append(_info.__dict__)

        result.print_table_raw(info_dicts, [], auto_header=True)

    else:
        result.print_result_as_table()



"""
FortiManager get ADOMs utility
"""

import logging
from typing import Any

from fotoobo.exceptions.exceptions import GeneralError
from fotoobo.helpers.config import config
from fotoobo.helpers.result import Result
from fotoobo.inventory import Inventory

log = logging.getLogger("fotoobo")


def adoms(host: str) -> Result[str]:
    """
    FortiManager get ADOMs

    Args:
        host: The host from the inventory to get the ADOMs list from

    Returns:
        List of ADOMs where ADOM is a dict with keys: name, version
    """
    inventory = Inventory(config.inventory_file)
    fmg = inventory.get_item(host, "fortimanager")
    result = Result[str]()
    log.debug("FortiManager get adoms ...")
    fmg.login()
    fmg_adoms = fmg.get_adoms()

    for adom in fmg_adoms:
        result.push_result(adom["name"], f"{adom['os_ver']}.{adom['mr']}")

    fmg.logout()
    return result


def devices(host: str) -> Result[dict[str, str | list[str]]]:
    """
    FortiManager get logical devices
    In a cluster only the cluster device is returned, not the physical nodes

    Args:
        host: The FortiManager from the inventory to get the Fortinet devices list from

    Returns:
        List of devices
    """
    inventory = Inventory(config.inventory_file)
    fmg = inventory.get_item(host, "fortimanager")
    log.debug("FortiManager get devices ...")
    fmg.login()
    payload = {
        "method": "get",
        "params": [
            {
                "loadsub": 1,
                "sortings": [{"build": 1}],
                "option": "object member",
                "url": "/dvmdb/device",
            }
        ],
    }
    response = fmg.api("post", payload=payload)
    fmg.logout()
    result = Result[dict[str, str | list[str]]]()

    for device in response.json()["result"][0]["data"]:
        data = {
            "version": f"{device['os_ver']}.{device['mr']}.{device['patch']}",
            "ha_mode": str(device["ha_mode"]),
            "platform": device["platform_str"],
            "desc": device["desc"],
            "ha_nodes": [],
        }
        if device["ha_mode"] == 1:
            ha_nodes = [node["name"] for node in device["ha_slave"]]
            data["ha_nodes"] = ha_nodes

        result.push_result(device["name"], data)

    return result


def policy(
    host: str, adom: str, policy_name: str, fields: list[str] | None = None
) -> Result[list[dict[str, Any]]]:
    """
    FortiManager get policy

    Args:
        host:
        adom:
        policy_name:
        fields:

    Returns:
        Result
    """
    fields = fields or [
        "status",
        # "_last_hit",  # data-format not clear
        "global-label",
        # "_hitcount",  # data-format not clear (it's not equal to the value in FortiManager)
        "policyid",
        "srcaddr",
        "groups",
        "dstaddr",
        "service",
        "action",
        "send-deny-packet",
        "comments",
    ]
    inventory = Inventory(config.inventory_file)
    fmg = inventory.get_item(host, "fortimanager")
    log.debug("FortiManager get policy '%s' from '%s' ...", policy_name, adom)
    fmg.login()
    payload = {
        "method": "get",
        "params": [
            {
                "option": "object member",
                "url": f"/pm/config/adom/{adom}/pkg/{policy_name}/firewall/policy",
            }
        ],
        "session": "",
        "id": 1,
    }
    result = fmg.api("post", payload=payload, timeout=30)
    data = result.json()

    if data["result"][0]["status"]["code"] != 0:
        code = data["result"][0]["status"]["code"]
        message = data["result"][0]["status"]["message"]
        log.error("FortiManager '%s' returned '%s': '%s'", host, code, message)
        raise GeneralError(f"FortiManager {host} returned {code}: {message}")

    policies = []
    for pol in data["result"][0]["data"]:
        policies.append({field: pol.get(field, None) for field in fields})

    fmg.logout()
    out_result = Result[list[dict[str, Any]]]()
    out_result.push_result(host, policies)
    return out_result


def version(host: str) -> Result[str]:
    """
    FortiManager get version

    Args:
        host: The host from the inventory to get the version

    Returns:
        Version data in a dict with keys: host, version
    """
    inventory = Inventory(config.inventory_file)
    fmg = inventory.get_item(host, "fortimanager")
    log.debug("FortiManager get version ...")
    fmg_version = fmg.get_version()
    result = Result[str]()
    result.push_result(host, fmg_version)
    return result



"""
The FortiGate commands
"""

import logging
from typing import Annotated

import typer

from fotoobo.helpers import cli_path
from fotoobo.tools import fgt

from .cmdb import cmdb

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
log = logging.getLogger("fotoobo")


app.add_typer(cmdb.app, name="cmdb", help="FortiGate get cmdb commands.")


@app.callback()
def callback(context: typer.Context) -> None:
    """
    The fgt get subcommand callback

    Args:
        context: The context object of the typer app
    """
    cli_path.append(str(context.invoked_subcommand))
    log.debug("About to execute command: '%s'", context.invoked_subcommand)


@app.command()
def version(
    host: Annotated[
        str | None,
        typer.Argument(
            help="The FortiGate hostname to access (must be defined in the inventory). ",
            metavar="[host]",
        ),
    ] = None,
) -> None:
    """
    Get the FortiGate(s) version(s).

    The optional argument \\[host] makes this command somewhat magic. If you omit \\[host] it
    searches for all devices of type 'fortigate' in the inventory and tries to get their FortiOS
    version.
    """
    result = fgt.get.version(host)
    result.print_result_as_table(title="FortiGate Versions", headers=["FortiGate", "Version"])
